"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSONHover = exports.jsonSchemaHover = void 0;
const json_schema_library_1 = require("json-schema-library");
const jsonPointers_1 = require("./utils/jsonPointers");
const formatting_1 = require("./utils/formatting");
const getSchema_1 = __importDefault(require("./utils/schema-lib/getSchema"));
const debug_1 = require("./utils/debug");
const dom_1 = require("./utils/dom");
/**
 * provides a JSON schema enabled tooltip extension for codemirror
 * @group Codemirror Extensions
 */
function jsonSchemaHover(schema, options) {
    const hover = new JSONHover(schema, options);
    return async function jsonDoHover(view, pos, side) {
        return hover.doHover(view, pos, side);
    };
}
exports.jsonSchemaHover = jsonSchemaHover;
function formatComplexType(schema, type, draft) {
    return `${type}: ${(0, formatting_1.joinWithOr)(schema[type].map((s) => {
        var _a;
        return (_a = s.type) !== null && _a !== void 0 ? _a : draft.resolveRef(s).type;
    }))}`;
}
class JSONHover {
    constructor(schema, opts) {
        this.opts = opts;
        this.schema = new json_schema_library_1.Draft04(schema);
        this.opts = Object.assign({ parser: JSON.parse }, this.opts);
    }
    getDataForCursor(view, pos, side) {
        const pointer = (0, jsonPointers_1.jsonPointerForPosition)(view.state, pos, side);
        let data = undefined;
        // TODO: use the AST tree to return the right hand, data so that we don't have to parse the doc
        try {
            data = this.opts.parser(view.state.doc.toString());
        }
        catch (_a) { }
        if (!pointer) {
            return null;
        }
        // if the data is valid, we can infer a type for complex types
        let subSchema = (0, getSchema_1.default)(this.schema, pointer, data);
        if (subSchema.type === "error" && data !== undefined) {
            // if the data is invalid, we won't get the type - try again without the data
            subSchema = (0, getSchema_1.default)(this.schema, pointer, undefined);
            if (subSchema.type === "error") {
                return { pointer };
            }
        }
        return { schema: subSchema, pointer };
    }
    formatMessage(texts) {
        const { message, typeInfo } = texts;
        if (message) {
            return (0, dom_1.el)("div", { class: "cm6-json-schema-hover" }, [
                (0, dom_1.el)("div", {
                    class: "cm6-json-schema-hover--description",
                    text: message,
                }),
                (0, dom_1.el)("div", { class: "cm6-json-schema-hover--code-wrapper" }, [
                    (0, dom_1.el)("code", { class: "cm6-json-schema-hover--code", text: typeInfo }),
                ]),
            ]);
        }
        return (0, dom_1.el)("div", { class: "cm6-json-schema-hover" }, [
            (0, dom_1.el)("div", { class: "cm6-json-schema-hover--code-wrapper" }, [
                (0, dom_1.el)("code", { class: "cm6-json-schema-hover--code", text: typeInfo }),
            ]),
        ]);
    }
    getHoverTexts(data, draft) {
        let typeInfo = null;
        let message = null;
        const { schema } = data;
        if (schema.oneOf) {
            typeInfo = formatComplexType(schema, "oneOf", draft);
        }
        if (schema.anyOf) {
            typeInfo = formatComplexType(schema, "anyOf", draft);
        }
        if (schema.allOf) {
            typeInfo = formatComplexType(schema, "allOf", draft);
        }
        if (schema.type) {
            typeInfo = Array.isArray(schema.type)
                ? (0, formatting_1.joinWithOr)(schema.type)
                : schema.type;
        }
        if (schema.description) {
            message = schema.description;
        }
        return { message, typeInfo };
    }
    // return hover state for the current json schema property
    async doHover(view, pos, side) {
        var _a, _b, _c, _d;
        const start = pos, end = pos;
        try {
            const cursorData = this.getDataForCursor(view, pos, side);
            // if we don't have a (sub)schema, we can't show anything
            if (!(cursorData === null || cursorData === void 0 ? void 0 : cursorData.schema))
                return null;
            const getHoverTexts = (_b = (_a = this.opts) === null || _a === void 0 ? void 0 : _a.getHoverTexts) !== null && _b !== void 0 ? _b : this.getHoverTexts;
            const hoverTexts = getHoverTexts(cursorData, this.schema);
            // allow users to override the hover
            const formatter = (_d = (_c = this.opts) === null || _c === void 0 ? void 0 : _c.formatHover) !== null && _d !== void 0 ? _d : this.formatMessage;
            const formattedDom = formatter(hoverTexts);
            return {
                pos: start,
                end,
                arrow: true,
                // to mimic similar modes for other editors
                // otherwise, it gets into a z-index battle with completion/etc
                above: true,
                create: (view) => {
                    return {
                        dom: formattedDom,
                    };
                },
            };
        }
        catch (err) {
            debug_1.debug.log(err);
            return null;
        }
    }
}
exports.JSONHover = JSONHover;
