import { Draft04 } from "json-schema-library";
import { getJSONSchema, schemaStateField } from "./state.js";
import { joinWithOr } from "./utils/formatting.js";
import { parseJSONDocumentState } from "./utils/parseJSONDocument.js";
import { el } from "./utils/dom.js";
// return an object path that matches with the json-source-map pointer
const getErrorPath = (error) => {
    var _a, _b, _c;
    // if a pointer is present, return without #
    if (((_a = error === null || error === void 0 ? void 0 : error.data) === null || _a === void 0 ? void 0 : _a.pointer) && ((_b = error === null || error === void 0 ? void 0 : error.data) === null || _b === void 0 ? void 0 : _b.pointer) !== "#") {
        return error.data.pointer.slice(1);
    }
    // return plain data.property if present
    if ((_c = error === null || error === void 0 ? void 0 : error.data) === null || _c === void 0 ? void 0 : _c.property) {
        return `/${error.data.property}`;
    }
    // else, return the empty pointer to represent the whole document
    return "";
};
export const handleRefresh = (vu) => {
    return (vu.startState.field(schemaStateField) !== vu.state.field(schemaStateField));
};
/**
 * Helper for simpler class instantiaton
 * @group Codemirror Extensions
 */
export function jsonSchemaLinter(options) {
    const validation = new JSONValidation(options);
    return (view) => {
        return validation.doValidation(view);
    };
}
export class JSONValidation {
    constructor(options) {
        this.schema = null;
        // rewrite the error message to be more human readable
        this.rewriteError = (error) => {
            var _a, _b, _c, _d, _e;
            const errorData = error === null || error === void 0 ? void 0 : error.data;
            const errors = errorData === null || errorData === void 0 ? void 0 : errorData.errors;
            if (error.code === "one-of-error" && (errors === null || errors === void 0 ? void 0 : errors.length)) {
                return `Expected one of ${joinWithOr(errors, (data) => data.data.expected)}`;
            }
            if (error.code === "type-error") {
                return `Expected <code>${((_a = error === null || error === void 0 ? void 0 : error.data) === null || _a === void 0 ? void 0 : _a.expected) && Array.isArray((_b = error === null || error === void 0 ? void 0 : error.data) === null || _b === void 0 ? void 0 : _b.expected)
                    ? joinWithOr((_c = error === null || error === void 0 ? void 0 : error.data) === null || _c === void 0 ? void 0 : _c.expected)
                    : (_d = error === null || error === void 0 ? void 0 : error.data) === null || _d === void 0 ? void 0 : _d.expected}</code> but received <code>${(_e = error === null || error === void 0 ? void 0 : error.data) === null || _e === void 0 ? void 0 : _e.received}</code>`;
            }
            const message = error.message
                // don't mention root object
                .replaceAll("in `#` ", "")
                .replaceAll("/", ".")
                .replaceAll("#.", "")
                // replace backticks with <code> tags
                .replaceAll(/`([^`]*)`/gm, "<code>$1</code>");
            return message;
        };
        this.options = Object.assign({ jsonParser: parseJSONDocumentState }, options);
        // TODO: support other versions of json schema.
        // most standard schemas are draft 4 for some reason, probably
        // backwards compatibility
        //
        // ajv did not support draft 4, so I used json-schema-library
    }
    get schemaTitle() {
        var _a, _b, _c;
        return (_c = (_b = (_a = this.schema) === null || _a === void 0 ? void 0 : _a.getSchema()) === null || _b === void 0 ? void 0 : _b.title) !== null && _c !== void 0 ? _c : "json-schema";
    }
    // validate using view as the linter extension signature requires
    doValidation(view) {
        const schema = getJSONSchema(view.state);
        if (!schema) {
            return [];
        }
        this.schema = new Draft04(schema);
        if (!this.schema)
            return [];
        const text = view.state.doc.toString();
        // ignore blank json strings
        if (!text || text.trim().length < 3)
            return [];
        const json = this.options.jsonParser(view.state);
        let errors = [];
        try {
            errors = this.schema.validate(json.data);
        }
        catch (_a) { }
        if (!errors.length)
            return [];
        // reduce() because we want to filter out errors that don't have a pointer
        return errors.reduce((acc, error) => {
            const errorPath = getErrorPath(error);
            const pointer = json.pointers.get(errorPath);
            if (pointer) {
                // if the error is a property error, use the key position
                const isPropertyError = error.name === "NoAdditionalPropertiesError";
                const errorString = this.rewriteError(error);
                acc.push({
                    from: isPropertyError ? pointer.keyFrom : pointer.valueFrom,
                    to: isPropertyError ? pointer.keyTo : pointer.valueTo,
                    message: errorString,
                    renderMessage: () => {
                        const dom = el("div", {});
                        dom.innerHTML = errorString;
                        return dom;
                    },
                    severity: "error",
                    source: this.schemaTitle,
                });
            }
            return acc;
        }, []);
    }
}
